# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
from unicore import utils
from unicore.models import BaseUnicoreModel, register_model, register_model_architecture
from unicore.modules import LayerNorm, init_bert_params
from .transformer_encoder_with_pair import TransformerEncoderWithPair
from typing import Dict, Any, List


logger = logging.getLogger(__name__)


@register_model("unimat")
class UniMatModel(BaseUnicoreModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        parser.add_argument(
            "--encoder-layers", type=int, metavar="L", help="num encoder layers"
        )
        parser.add_argument(
            "--encoder-embed-dim",
            type=int,
            metavar="H",
            help="encoder embedding dimension",
        )
        parser.add_argument(
            "--encoder-ffn-embed-dim",
            type=int,
            metavar="F",
            help="encoder embedding dimension for FFN",
        )
        parser.add_argument(
            "--encoder-attention-heads",
            type=int,
            metavar="A",
            help="num encoder attention heads",
        )
        parser.add_argument(
            "--activation-fn",
            choices=utils.get_available_activation_fns(),
            help="activation function to use",
        )
        parser.add_argument(
            "--pooler-activation-fn",
            choices=utils.get_available_activation_fns(),
            help="activation function to use for pooler layer",
        )
        parser.add_argument(
            "--emb-dropout", type=float, metavar="D", help="dropout probability for embeddings"
        )
        parser.add_argument(
            "--dropout", type=float, metavar="D", help="dropout probability"
        )
        parser.add_argument(
            "--attention-dropout",
            type=float,
            metavar="D",
            help="dropout probability for attention weights",
        )
        parser.add_argument(
            "--activation-dropout",
            type=float,
            metavar="D",
            help="dropout probability after activation in FFN",
        )
        parser.add_argument(
            "--pooler-dropout",
            type=float,
            metavar="D",
            help="dropout probability in the masked_lm pooler layers",
        )
        parser.add_argument(
            "--max-seq-len", type=int, help="number of positional embeddings to learn"
        )
        parser.add_argument(
            "--post-ln", type=bool, help="use post layernorm or pre layernorm"
        )
        parser.add_argument(
            "--masked-token-loss",
            type=float,
            metavar="D",
            help="mask loss ratio",
        )
        parser.add_argument(
            "--masked-dist-loss",
            type=float,
            metavar="D",
            help="masked distance loss ratio",
        )
        parser.add_argument(
            "--masked-coord-loss",
            type=float,
            metavar="D",
            help="masked coord loss ratio",
        )
        parser.add_argument(
            "--x-norm-loss",
            type=float,
            metavar="D",
            help="x norm loss ratio",
        )
        parser.add_argument(
            "--delta-pair-repr-norm-loss",
            type=float,
            metavar="D",
            help="delta encoder pair repr norm loss ratio",
        )
        parser.add_argument(
            "--masked-coord-dist-loss",
            type=float,
            metavar="D",
            help="masked coord dist loss ratio",
        )
        parser.add_argument(
            "--lattice-loss",
            type=float,
            metavar="D",
            help="lattice loss ratio",
        )
        parser.add_argument(
            "--gaussian-kernel",
            action="store_true",
            help="use gaussian kernel for distance map",
        )
        # parser.add_argument(
        #     "--atom-descriptor",
        #     action="store_true",
        #     help="use extra atom descriptor",
        # )  
        # parser.add_argument(
        #     "--global-distance",
        #     action="store_true",
        #     help="use global distance",
        # ) 

    def __init__(self, args, dictionary):
        super().__init__()
        base_architecture(args)
        self.args = args
        self.padding_idx = dictionary.pad()
        self.embed_tokens = nn.Embedding(len(dictionary), args.encoder_embed_dim, self.padding_idx)
        self._num_updates = None
        self.encoder = TransformerEncoderWithPair(
            encoder_layers=args.encoder_layers,
            embed_dim=args.encoder_embed_dim,
            ffn_embed_dim=args.encoder_ffn_embed_dim,
            attention_heads=args.encoder_attention_heads,
            emb_dropout=args.emb_dropout,
            dropout=args.dropout,
            attention_dropout=args.attention_dropout,
            activation_dropout=args.activation_dropout,
            max_seq_len=args.max_seq_len,
            activation_fn=args.activation_fn,
            no_final_head_layer_norm=args.delta_pair_repr_norm_loss < 0,
        )
        if args.masked_token_loss > 0:
            self.lm_head = MaskLMHead(embed_dim=args.encoder_embed_dim,
                output_dim=len(dictionary),
                activation_fn=args.activation_fn,
                weight=self.embed_tokens.weight,
            )

        K = 128
        n_edge_type = len(dictionary) * len(dictionary)
        self.gbf_proj = NonLinearHead(K, args.encoder_attention_heads, args.activation_fn)
        if 'gaussian_kernel' in args and args.gaussian_kernel:
            if args.global_distance:
                self.gbf = GlobalGaussianLayer(K, n_edge_type)
            else:
                self.gbf = GaussianLayer(K, n_edge_type)
        else:
            self.gbf = NumericalEmbed(K, n_edge_type)
        if args.masked_coord_loss > 0:
            self.pair2coord_proj = NonLinearHead(args.encoder_attention_heads, 1, args.activation_fn)
        if args.masked_dist_loss > 0:
            if args.global_distance:
                self.glodist_head = GlobalDistanceHead(args.encoder_attention_heads, args.activation_fn)
            else:
                self.dist_head = DistanceHead(args.encoder_attention_heads, args.activation_fn)
        if args.lattice_loss > 0:
            # self.lattice_head = ClassificationHead(input_dim=args.encoder_embed_dim,
            #                                     inner_dim=args.encoder_embed_dim,
            #                                     num_classes=3,
            #                                     activation_fn=self.args.pooler_activation_fn,
            #                                     pooler_dropout=self.args.pooler_dropout,
            # )
            self.lattice_head = NodeClassificationHead(input_dim=args.encoder_embed_dim,
                                                inner_dim=args.encoder_embed_dim,
                                                num_classes=6,
                                                activation_fn=self.args.pooler_activation_fn,
                                                pooler_dropout=self.args.pooler_dropout,
            )
        self.classification_heads = nn.ModuleDict()
        self.node_classification_heads = nn.ModuleDict()
        self.apply(init_bert_params)

    @classmethod
    def build_model(cls, args, task):
        """Build a new model instance."""
        return cls(args, task.dictionary)

    def forward(
        self,
        select_atom,
        src_tokens,
        src_distance,
        src_coord,
        src_edge_type,
        atom_descriptor=None,
        encoder_masked_tokens=None,
        features_only=False,
        classification_head_name=None,
        **kwargs
    ):

        if classification_head_name is not None:
            features_only = True

        padding_mask = src_tokens.eq(self.padding_idx)
        if not padding_mask.any():
            padding_mask = None
        x = self.embed_tokens(src_tokens)
        def get_dist_features(dist, et):
            n_node = et.size(-1)
            gbf_feature = self.gbf(dist, et)
            # print("gbf_feature", gbf_feature.shape)
            gbf_result = self.gbf_proj(gbf_feature)
            graph_attn_bias = gbf_result
            # print("gbf_result", gbf_result.shape)
            graph_attn_bias = graph_attn_bias.permute(0, 3, 1, 2).contiguous()
            # print("graph_attn_bias", graph_attn_bias.shape)
            graph_attn_bias = graph_attn_bias.view(-1, n_node, n_node)
            # print("graph_attn_bias", graph_attn_bias.shape)
            return graph_attn_bias
        # print(x.shape, src_distance.shape, src_edge_type.shape)
        graph_attn_bias = get_dist_features(src_distance, src_edge_type)
        encoder_rep, encoder_pair_rep, delta_encoder_pair_rep, x_norm, delta_encoder_pair_rep_norm = self.encoder(x, padding_mask=padding_mask, attn_mask=graph_attn_bias)
        encoder_pair_rep[encoder_pair_rep == float('-inf')] = 0
        # print(encoder_rep.shape, graph_attn_bias.shape, encoder_pair_rep.shape, delta_encoder_pair_rep.shape, x_norm, delta_encoder_pair_rep_norm)
        encoder_distance = None
        encoder_coord = None
        lattice = None

        if not features_only:
            if self.args.masked_token_loss > 0:
                logits = self.lm_head(encoder_rep, encoder_masked_tokens)
            if self.args.masked_coord_loss > 0:
                coords_emb = src_coord
                if padding_mask is not None:
                    atom_num = (torch.sum(1-padding_mask.type_as(x), dim=1) - 1).view(-1, 1, 1, 1)
                else:
                    atom_num = src_coord.shape[1] - 1
                delta_pos = coords_emb.unsqueeze(1) - coords_emb.unsqueeze(2)
                attn_probs = self.pair2coord_proj(delta_encoder_pair_rep)
                coord_update = delta_pos / atom_num * attn_probs
                coord_update = torch.sum(coord_update, dim=2)
                encoder_coord = coords_emb + coord_update
            if self.args.masked_dist_loss > 0:
                if self.args.global_distance:
                    encoder_distance = self.glodist_head(encoder_pair_rep)
                else:
                    encoder_distance = self.dist_head(encoder_pair_rep)
            if self.args.lattice_loss > 0:
                lattice = self.lattice_head(encoder_rep[select_atom==1])
        if atom_descriptor is not None:
            model_input = torch.cat((encoder_rep, atom_descriptor.to(encoder_rep.dtype)), dim=2)
        else : 
            model_input = encoder_rep
        if classification_head_name is not None:
            if classification_head_name in self.classification_heads:
                logits = self.classification_heads[classification_head_name](model_input)
            elif classification_head_name in self.node_classification_heads:
                logits = self.node_classification_heads[classification_head_name](model_input[select_atom==1])
        elif features_only and (self.classification_heads or self.node_classification_heads):
            logits = {}
            for name, head in self.node_classification_heads.items():
                logits[name] = head(model_input[select_atom==1])
            for name, head in self.classification_heads.items():
                logits[name] = head(model_input)

        return logits, encoder_distance, encoder_coord, lattice, x_norm, delta_encoder_pair_rep_norm

    def register_classification_head(
        self, name, num_classes=None, inner_dim=None, extra_dim=0, **kwargs
    ):
        """Register a classification head."""
        if name in self.classification_heads:
            prev_num_classes = self.classification_heads[name].out_proj.out_features
            prev_inner_dim = self.classification_heads[name].dense.out_features
            if num_classes != prev_num_classes or inner_dim != prev_inner_dim:
                logger.warning(
                    're-registering head "{}" with num_classes {} (prev: {}) '
                    "and inner_dim {} (prev: {})".format(
                        name, num_classes, prev_num_classes, inner_dim, prev_inner_dim
                    )
                )
        input_dim=self.args.encoder_embed_dim + extra_dim
        self.classification_heads[name] = ClassificationHead(
            input_dim=input_dim,
            inner_dim=inner_dim or input_dim,
            num_classes=num_classes,
            activation_fn=self.args.pooler_activation_fn,
            pooler_dropout=self.args.pooler_dropout,
        )

    def register_node_classification_head(
        self, name, num_classes=None, inner_dim=None, extra_dim=0, **kwargs
    ):
        """Register a classification head."""
        if name in self.node_classification_heads:
            prev_num_classes = self.node_classification_heads[name].out_proj.out_features
            prev_inner_dim = self.node_classification_heads[name].dense.out_features
            if num_classes != prev_num_classes or inner_dim != prev_inner_dim:
                logger.warning(
                    're-registering head "{}" with num_classes {} (prev: {}) '
                    "and inner_dim {} (prev: {})".format(
                        name, num_classes, prev_num_classes, inner_dim, prev_inner_dim
                    )
                )
        input_dim=self.args.encoder_embed_dim + extra_dim
        self.node_classification_heads[name] = NodeClassificationHead(
            input_dim=input_dim,
            inner_dim=inner_dim or input_dim,
            num_classes=num_classes,
            activation_fn=self.args.pooler_activation_fn,
            pooler_dropout=self.args.pooler_dropout,
        )    

    def set_num_updates(self, num_updates):
        """State from trainer to pass along to model at every update."""
        self._num_updates = num_updates

    def get_num_updates(self):
        return self._num_updates


class MaskLMHead(nn.Module):
    """Head for masked language modeling."""

    def __init__(self, embed_dim, output_dim, activation_fn, weight=None):
        super().__init__()
        self.dense = nn.Linear(embed_dim, embed_dim)
        self.activation_fn = utils.get_activation_fn(activation_fn)
        self.layer_norm = LayerNorm(embed_dim)

        if weight is None:
            weight = nn.Linear(embed_dim, output_dim, bias=False).weight
        self.weight = weight
        self.bias = nn.Parameter(torch.zeros(output_dim))

    def forward(self, features, masked_tokens=None, **kwargs):
        # Only project the masked tokens while training,
        # saves both memory and computation
        if masked_tokens is not None:
            features = features[masked_tokens, :]

        x = self.dense(features)
        x = self.activation_fn(x)
        x = self.layer_norm(x)
        # project back to size of vocabulary with bias
        x = F.linear(x, self.weight) + self.bias
        return x


class ClassificationHead(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(
        self,
        input_dim,
        inner_dim,
        num_classes,
        activation_fn,
        pooler_dropout,
    ):
        super().__init__()
        self.dense = nn.Linear(input_dim, inner_dim)
        self.activation_fn = utils.get_activation_fn(activation_fn)
        self.dropout = nn.Dropout(p=pooler_dropout)
        self.out_proj = nn.Linear(inner_dim, num_classes)

    def forward(self, features, **kwargs):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout(x)
        x = self.dense(x)
        x = self.activation_fn(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x

class NodeClassificationHead(nn.Module):
    """Head for node-level classification tasks."""

    def __init__(
        self,
        input_dim,
        inner_dim,
        num_classes,
        activation_fn,
        pooler_dropout,
    ):
        super().__init__()
        self.dense = nn.Linear(input_dim, inner_dim)
        self.activation_fn = utils.get_activation_fn(activation_fn)
        self.dropout = nn.Dropout(p=pooler_dropout)
        self.out_proj = nn.Linear(inner_dim, num_classes)

    def forward(self, features, **kwargs):
        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = self.activation_fn(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x


class NonLinearHead(nn.Module):
    """Head for simple classification tasks."""

    def __init__(
        self,
        input_dim,
        out_dim,
        activation_fn,
        hidden=None,
    ):
        super().__init__()
        hidden = input_dim if not hidden else hidden
        self.linear1 = nn.Linear(input_dim, hidden)
        self.linear2 = nn.Linear(hidden, out_dim)
        self.activation_fn = utils.get_activation_fn(activation_fn)

    def forward(self, x):
        x = self.linear1(x)
        x = self.activation_fn(x)
        x = self.linear2(x)
        return x


class DistanceHead(nn.Module):
    def __init__(
        self,
        heads,
        activation_fn,
    ):
        super().__init__()
        self.dense = nn.Linear(heads, heads)
        self.layer_norm = nn.LayerNorm(heads)
        self.out_proj = nn.Linear(heads, 1)
        self.activation_fn = utils.get_activation_fn(activation_fn)

    def forward(self, x):
        bsz, seq_len, seq_len, _ = x.size()
        # x[x == float('-inf')] = 0
        x = self.dense(x)
        x = self.activation_fn(x)
        x = self.layer_norm(x)
        x = self.out_proj(x).view(bsz, seq_len, seq_len)
        x = (x + x.transpose(-1, -2)) * 0.5
        return x

class GlobalDistanceHead(nn.Module):
    def __init__(
        self,
        heads,
        activation_fn,
    ):
        super().__init__()
        self.dense = nn.Linear(heads, heads)
        self.layer_norm = nn.LayerNorm(heads)
        self.out_proj = nn.Linear(heads, 4)
        self.activation_fn = utils.get_activation_fn(activation_fn)

    def forward(self, x):
        bsz, seq_len, seq_len, _ = x.size()
        # x[x == float('-inf')] = 0
        x = self.dense(x)
        x = self.activation_fn(x)
        x = self.layer_norm(x)
        x = self.out_proj(x)
        x = (x + x.transpose(1, 2)) * 0.5
        return x

@torch.jit.script
def gaussian(x, mean, std):
    pi = 3.14159
    a = (2*pi) ** 0.5
    return torch.exp(-0.5 * (((x - mean) / std) ** 2)) / (a * std)


class GaussianLayer(nn.Module):
    def __init__(self, K=128, edge_types=1024):
        super().__init__()
        self.K = K
        self.means = nn.Embedding(1, K)
        self.stds = nn.Embedding(1, K)
        self.mul = nn.Embedding(edge_types, 1)
        self.bias = nn.Embedding(edge_types, 1)
        nn.init.uniform_(self.means.weight, 0, 3)
        nn.init.uniform_(self.stds.weight, 0, 3)
        nn.init.constant_(self.bias.weight, 0)
        nn.init.constant_(self.mul.weight, 1)

    def forward(self, x, edge_type):  # x: (atoms, hidden),  edge_type: atoms, atoms
        mul = self.mul(edge_type).type_as(x)
        bias = self.bias(edge_type).type_as(x)
        x = mul * x.unsqueeze(-1) + bias
        x = x.expand(-1, -1, -1, self.K)
        mean = self.means.weight.float().view(-1)
        std = self.stds.weight.float().view(-1).abs() + 1e-5
        return gaussian(x.float(), mean, std).type_as(self.means.weight)

class GlobalGaussianLayer(nn.Module):
    def __init__(self, K=128, edge_types=1024):
        super().__init__()
        self.K = K
        self.means = nn.Embedding(1, K)
        self.stds = nn.Embedding(1, K)
        self.mul = nn.Embedding(edge_types, 1)
        self.bias = nn.Embedding(edge_types, 1)
        nn.init.uniform_(self.means.weight, 0, 3)
        nn.init.uniform_(self.stds.weight, 0, 3)
        nn.init.constant_(self.bias.weight, 0)
        nn.init.constant_(self.mul.weight, 1)

    def forward(self, x, edge_type):  # x: (atoms, hidden),  edge_type: atoms, atoms
        mul = self.mul(edge_type).type_as(x)
        bias = self.bias(edge_type).type_as(x)
        x = mul * x + bias     # (bsz, n, n, 16)
        # x = x.repeat(1, 1, 1, int(self.K/x.shape[-1]))
        # replication_rules = [18, 14, 12, 12, 12, 12, 8, 8, 8, 8, 3, 3, 3, 3, 2, 2]
        # replication_rules = [66, 8, 8, 8, 8, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2]
        # replication_rules = [30, 17, 17, 17, 17, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2]
        # replication_rules = [80, 32, 10, 6]
        replication_rules = [128, 0, 0, 0]
        # replication_rules = [88, 40, 0, 0]
        x = torch.repeat_interleave(x, torch.tensor(replication_rules).to(x.device), dim=-1)  # (bsz, n, n, 128)
        mean = self.means.weight.float().view(-1)
        std = self.stds.weight.float().view(-1).abs() + 1e-5
        return gaussian(x.float(), mean, std).type_as(self.means.weight)

class NumericalEmbed(nn.Module):
    def __init__(self, K=128, edge_types=1024, activation_fn='gelu'):
        super().__init__()
        self.K = K 
        self.mul = nn.Embedding(edge_types, 1)
        self.bias = nn.Embedding(edge_types, 1)
        self.w_edge = nn.Embedding(edge_types, K)

        self.proj = NonLinearHead(1, K, activation_fn, hidden=2*K)
        self.ln = LayerNorm(K)

        nn.init.constant_(self.bias.weight, 0)
        nn.init.constant_(self.mul.weight, 1)
        nn.init.kaiming_normal_(self.w_edge.weight)


    def forward(self, x, edge_type):    # edge_type, atoms
        x = x[:,:,:,0].squeeze(-1)
        mul = self.mul(edge_type).type_as(x)
        bias = self.bias(edge_type).type_as(x)
        w_edge = self.w_edge(edge_type).type_as(x)
        edge_emb = w_edge * torch.sigmoid(mul * x.unsqueeze(-1) + bias)
        
        edge_proj = x.unsqueeze(-1).type_as(self.mul.weight)
        edge_proj = self.proj(edge_proj)
        edge_proj = self.ln(edge_proj)

        h = edge_proj + edge_emb
        h = h.type_as(self.mul.weight)
        return h
        

@register_model_architecture("unimat", "unimat")
def base_architecture(args):
    args.encoder_layers = getattr(args, "encoder_layers", 8)
    args.encoder_embed_dim = getattr(args, "encoder_embed_dim", 512)
    args.encoder_ffn_embed_dim = getattr(args, "encoder_ffn_embed_dim", 2048)
    args.encoder_attention_heads = getattr(args, "encoder_attention_heads", 64)
    args.dropout = getattr(args, "dropout", 0.1)
    args.emb_dropout = getattr(args, "emb_dropout", 0.1)
    args.attention_dropout = getattr(args, "attention_dropout", 0.1)
    args.activation_dropout = getattr(args, "activation_dropout", 0.0)
    args.pooler_dropout = getattr(args, "pooler_dropout", 0.0)
    args.max_seq_len = getattr(args, "max_seq_len", 1024)
    args.activation_fn = getattr(args, "activation_fn", "gelu")
    args.pooler_activation_fn = getattr(args, "pooler_activation_fn", "tanh")
    args.post_ln = getattr(args, "post_ln", False)
    args.masked_token_loss = getattr(args, "masked_token_loss", -1.0)
    args.masked_coord_loss = getattr(args, "masked_coord_loss", -1.0)
    args.masked_dist_loss = getattr(args, "masked_dist_loss", -1.0)
    args.x_norm_loss = getattr(args, "x_norm_loss", -1.0)
    args.delta_pair_repr_norm_loss = getattr(args, "delta_pair_repr_norm_loss", -1.0)
    args.lattice_loss = getattr(args, "lattice_loss", -1.0)


@register_model_architecture("unimat", "unimat_base")
def unimol_base_architecture(args):
    base_architecture(args)

@register_model_architecture("unimat", "unimol_large")
def unimol_base_architecture(args):
    base_architecture(args)
    args.encoder_layers = 15



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os

import numpy as np
import torch
from unicore.data import (
    Dictionary,
    NestedDictionaryDataset,
    AppendTokenDataset,
    PrependTokenDataset,
    RightPadDataset,
    EpochShuffleDataset,
    TokenizeDataset,
    RightPadDataset2D,
    LMDBDataset,
    RawLabelDataset,
)
from ..data import (
    KeyDataset,
    ConformerSampleDataset,
    TTADataset,
    IndexDataset,
    TTAIndexDataset,
    ToTorchDataset,
    MaskPointsDataset,
    DistanceDataset,
    GlobalDistanceDataset,
    EdgeTypeDataset,
    RightPadDataset3D,
    PrependAndAppend2DDataset,
    PrependAndAppend3DDataset,
    RightPadDataset2D0,
    LatticeNormalizeDataset,
    LatticeMatrixNormalizeDataset,
    RemoveHydrogenDataset,
    CroppingDataset,
    NormalizeDataset,
    TargetScalerDataset,
    FoldLMDBDataset,
    StackedLMDBDataset,
    SplitLMDBDataset,
    SelectTokenDataset,
    FilterDataset,
)
from unicore.tasks import UnicoreTask, register_task
from uninmr.utils import parse_select_atom, TargetScaler
from ase import Atoms
logger = logging.getLogger(__name__)

@register_task("uninmr")
class UniNMRTask(UnicoreTask):
    """Task for training transformer auto-encoder models."""

    @staticmethod
    def add_args(parser):
        """Add task-specific arguments to the parser."""
        parser.add_argument(
            "data",
            help="downstream data path"
        )
        parser.add_argument(
            "--saved-dir",
            help="saved dir"
        )
        parser.add_argument(
            "--classification-head-name",
            default="nmr_head",
            help="finetune downstream task name"
        )
        parser.add_argument(
            "--num-classes",
            default=1,
            type=int,
            help="finetune downstream task classes numbers"
        )
        parser.add_argument(
            "--max-atoms",
            type=int,
            default=512,
            help="selected maximum number of atoms in a molecule",
        )
        parser.add_argument(
            "--dict-name",
            default="dict.txt",
            help="dictionary file",
        )
        parser.add_argument(
            "--remove-hydrogen",
            action="store_true",
            help="remove hydrogen atoms",
        )
        parser.add_argument(
            "--has-matid",
            action="store_true",
            help="whether already has matid",
        )
        parser.add_argument(
            "--conformer-augmentation",
            action="store_true",
            help="using conformer augmentation",
        )
        parser.add_argument(
            "--conf-size",
            type=int,
            default=10,
            help="conformer nums per structure",
        )
        parser.add_argument(
            "--global-distance",
            action="store_true",
            help="use global distance",
        )
        parser.add_argument(
            "--atom-descriptor",
            type=int,
            default=0,
            help="use extra atom descriptor",
        )
        parser.add_argument(
            "--selected-atom",
            default="All",
            help="select atom: All or H or H&C&F...",
        )
        parser.add_argument(
            '--split-mode',
            type=str,
            default='predefine',
            choices=['predefine', 'cross_valid', 'random', 'infer'],
        )
        parser.add_argument(
            "--nfolds",
            default=5,
            type=int,
            help="cross validation split folds"
        )
        parser.add_argument(
            "--fold",
            default=0,
            type=int,
            help='local fold used as validation set, and other folds will be used as train set'
        )
        parser.add_argument(
            "--cv-seed",
            default=42,
            type=int,
            help="random seed used to do cross validation splits"
        )

    def __init__(self, args, dictionary):
        super().__init__(args)
        self.dictionary = dictionary
        self.seed = args.seed
        # add mask token
        self.mask_idx = dictionary.add_symbol("[MASK]", is_special=True)
        self.selected_token = parse_select_atom(self.dictionary, args.selected_atom)
        if args.saved_dir == None:
            self.args.saved_dir = args.save_dir
        self.target_scaler = TargetScaler(args.saved_dir)

        if self.args.split_mode =='predefine':
            train_path = os.path.join(self.args.data, "train" + ".lmdb")
            self.train_dataset = LMDBDataset(train_path)
            valid_path = os.path.join(self.args.data, "valid" + ".lmdb")
            self.valid_dataset = LMDBDataset(valid_path)
            atoms_target = np.concatenate([np.array(self.train_dataset[i]['atoms_target']) for i in range(len(self.train_dataset))], axis=0)
            atoms_target_mask = np.concatenate([np.array(self.train_dataset[i]['atoms_target_mask']) for i in range(len(self.train_dataset))], axis=0)
            self.target_scaler.fit(target=atoms_target[atoms_target_mask==1].reshape(-1, self.args.num_classes), num_classes=self.args.num_classes, dump_dir=self.args.save_dir)
        elif self.args.split_mode == 'infer':
            valid_path = os.path.join(self.args.data, "valid" + ".lmdb")
            self.valid_dataset = LMDBDataset(valid_path)
        else:
            self.__init_data()

    def __init_data(self):
        data_path = os.path.join(self.args.data, 'train.lmdb')
        raw_dataset = LMDBDataset(data_path)
        atoms_target = np.concatenate([np.array(raw_dataset[i]['atoms_target']) for i in range(len(raw_dataset))], axis=0)
        atoms_target_mask = np.concatenate([np.array(raw_dataset[i]['atoms_target_mask']) for i in range(len(raw_dataset))], axis=0)

        if self.args.split_mode == 'cross_valid':
            train_folds = []
            for _fold in range(self.args.nfolds):
                if _fold == 0:
                    parent_dir = os.path.dirname(self.args.saved_dir)
                    self.target_scaler.fit(target=atoms_target[atoms_target_mask==1].reshape(-1, self.args.num_classes), num_classes=self.args.num_classes, dump_dir=parent_dir)
                    cache_fold_info = FoldLMDBDataset(raw_dataset, self.args.cv_seed, _fold, nfolds=self.args.nfolds).get_fold_info()
                if _fold == self.args.fold:
                    self.valid_dataset = FoldLMDBDataset(raw_dataset, self.args.cv_seed, _fold, nfolds=self.args.nfolds, cache_fold_info=cache_fold_info)
                if _fold != self.args.fold:
                    train_folds.append(FoldLMDBDataset(raw_dataset, self.args.cv_seed, _fold, nfolds=self.args.nfolds, cache_fold_info=cache_fold_info))
            self.train_dataset = StackedLMDBDataset(train_folds)
        elif self.args.split_mode == 'random':
            self.target_scaler.fit(target=atoms_target[atoms_target_mask==1].reshape(-1, self.args.num_classes), num_classes=self.args.num_classes, dump_dir=self.args.saved_dir)
            cache_fold_info = SplitLMDBDataset(raw_dataset, self.args.seed, 0).get_fold_info()
            self.train_dataset = SplitLMDBDataset(raw_dataset, self.args.seed, 0, cache_fold_info=cache_fold_info)
            self.valid_dataset = SplitLMDBDataset(raw_dataset, self.args.seed, 1, cache_fold_info=cache_fold_info)

    @classmethod
    def setup_task(cls, args, **kwargs):
        dictionary = Dictionary.load(os.path.join(args.data, args.dict_name))
        logger.info("dictionary: {} types".format(len(dictionary)))
        return cls(args, dictionary)

    def load_dataset(self, split, **kwargs):
        """Load a given dataset split.
        Args:
            split (str): name of the data scoure (e.g., train)
        """
        # self.split = split
        # if self.args.split_mode != 'predefine':
        if split == 'train':
            dataset = self.train_dataset
        elif split == 'valid':
            dataset =self.valid_dataset
        # else:
        #     split_path = os.path.join(self.args.data, split + ".lmdb")
        #     dataset = LMDBDataset(split_path)
        #     if split == 'train':
        #         atoms_target = np.concatenate([np.array(dataset[i]['atoms_target']) for i in range(len(dataset))], axis=0)
        #         atoms_target_mask = np.concatenate([np.array(dataset[i]['atoms_target_mask']) for i in range(len(dataset))], axis=0)
        #         self.target_scaler.fit(target=atoms_target[atoms_target_mask==1].reshape(-1, self.args.num_classes), num_classes=self.args.num_classes, dump_dir=self.args.save_dir)

        if self.args.has_matid:
            matid_dataset = KeyDataset(dataset, "matid")
        else:
            matid_dataset = IndexDataset(dataset)

        if self.args.conformer_augmentation:
            if split == 'train':
                dataset = ConformerSampleDataset(dataset, self.seed, "atoms", "coordinates_list")
            else:
                dataset = TTADataset(dataset, self.seed, "atoms", "coordinates_list", self.args.conf_size)
                matid_dataset = TTAIndexDataset(matid_dataset, self.args.conf_size)
        if self.args.remove_hydrogen:
            dataset = RemoveHydrogenDataset(dataset, "atoms", "coordinates")
        dataset = CroppingDataset(dataset, self.seed, "atoms", "coordinates", self.args.max_atoms)
        dataset = NormalizeDataset(dataset, "coordinates")

        # lattice_dataset = LatticeNormalizeDataset(dataset, 'abc', 'angles')
        # lattice_dataset = ToTorchDataset(lattice_dataset, 'float32')

        token_dataset = KeyDataset(dataset, "atoms")
        token_dataset = TokenizeDataset(token_dataset, self.dictionary, max_seq_len=self.args.max_seq_len)
        atoms_target_mask_dataset = KeyDataset(dataset, "atoms_target_mask")
        select_atom_dataset = SelectTokenDataset(token_dataset=token_dataset, token_mask_dataset=atoms_target_mask_dataset, selected_token=self.selected_token)
        filter_list = [0 if torch.all(select_atom_dataset[i]==0) else 1 for i in range(len(select_atom_dataset))]

        dataset = FilterDataset(dataset, filter_list)
        matid_dataset = FilterDataset(matid_dataset, filter_list)
        token_dataset = FilterDataset(token_dataset, filter_list)
        select_atom_dataset = FilterDataset(select_atom_dataset, filter_list)

        coord_dataset = KeyDataset(dataset, "coordinates")

        def PrependAndAppend(dataset, pre_token, app_token):
            dataset = PrependTokenDataset(dataset, pre_token)
            return AppendTokenDataset(dataset, app_token)

        token_dataset = PrependAndAppend(token_dataset, self.dictionary.bos(), self.dictionary.eos())
        select_atom_dataset = PrependAndAppend(select_atom_dataset, self.dictionary.pad(), self.dictionary.pad())

        coord_dataset = ToTorchDataset(coord_dataset, 'float32')

        if self.args.global_distance:
            lattice_matrix_dataset = LatticeMatrixNormalizeDataset(dataset, 'lattice_matrix')
            logger.info("use global distance: {}".format(self.args.global_distance))
            distance_dataset = GlobalDistanceDataset(coord_dataset, lattice_matrix_dataset)
            distance_dataset = PrependAndAppend3DDataset(distance_dataset, 0.0)
            distance_dataset = RightPadDataset3D(distance_dataset, pad_idx=0)
        else:
            distance_dataset = DistanceDataset(coord_dataset)
            distance_dataset = PrependAndAppend2DDataset(distance_dataset, 0.0)
            distance_dataset = RightPadDataset2D(distance_dataset, pad_idx=0)
        coord_dataset = PrependAndAppend(coord_dataset, 0.0, 0.0)
        edge_type = EdgeTypeDataset(token_dataset, len(self.dictionary))

        tgt_dataset = KeyDataset(dataset, "atoms_target")
        tgt_dataset = TargetScalerDataset(tgt_dataset, self.target_scaler, self.args.num_classes)
        tgt_dataset = ToTorchDataset(tgt_dataset, dtype='float32')

        tgt_dataset = PrependAndAppend(tgt_dataset, self.dictionary.pad(), self.dictionary.pad())

        if self.args.atom_descriptor != 0:
            atomdes_dataset = KeyDataset(dataset, "atoms_descriptor")
            atomdes_dataset = ToTorchDataset(atomdes_dataset, dtype='float32')
            atomdes_dataset = PrependAndAppend(atomdes_dataset, self.dictionary.pad(), self.dictionary.pad())
            nest_dataset = NestedDictionaryDataset(
                    {
                        "net_input": {
                            "select_atom": RightPadDataset(
                                select_atom_dataset,
                                pad_idx=self.dictionary.pad(),
                            ),
                            "src_tokens": RightPadDataset(
                                token_dataset,
                                pad_idx=self.dictionary.pad(),
                            ),
                            "src_coord": RightPadDataset2D0(
                                coord_dataset,
                                pad_idx=0,
                            ),
                            "src_distance": distance_dataset,
                            "src_edge_type": RightPadDataset2D(
                                edge_type,
                                pad_idx=0,
                            ),
                            "atom_descriptor": RightPadDataset2D0(
                                atomdes_dataset,
                                pad_idx=0,
                            ),
                        },
                        "target": {
                            "finetune_target": RightPadDataset(
                                tgt_dataset,
                                pad_idx=0,
                            ),
                        },
                        "matid": matid_dataset,
                    },
                )
        else:
            nest_dataset = NestedDictionaryDataset(
                    {
                        "net_input": {
                            "select_atom": RightPadDataset(
                                select_atom_dataset,
                                pad_idx=self.dictionary.pad(),
                            ),
                            "src_tokens": RightPadDataset(
                                token_dataset,
                                pad_idx=self.dictionary.pad(),
                            ),
                            "src_coord": RightPadDataset2D0(
                                coord_dataset,
                                pad_idx=0,
                            ),
                            "src_distance": distance_dataset,
                            "src_edge_type": RightPadDataset2D(
                                edge_type,
                                pad_idx=0,
                            ),
                        },
                        "target": {
                            "finetune_target": RightPadDataset(
                                tgt_dataset,
                                pad_idx=0,
                            ),
                        },
                        "matid": matid_dataset,
                    },
                )
        if split in ["train", "train.small"]:
            nest_dataset = EpochShuffleDataset(nest_dataset, len(nest_dataset), self.args.seed)
        self.datasets[split] = nest_dataset

    def build_model(self, args):
        from unicore import models
        model = models.build_model(args, self)
        model.register_node_classification_head(
            self.args.classification_head_name,
            num_classes=self.args.num_classes,
            extra_dim=self.args.atom_descriptor,
        )
        return model



import lmdb
import os
from tqdm import tqdm
import pickle
import numpy as np

def write_lmdb(outputfilename, nmr_data, nthreads=4):

    try:
        os.remove(outputfilename)
    except:
        pass
    env_new = lmdb.open(
        outputfilename,
        subdir=False,
        readonly=False,
        lock=False,
        readahead=False,
        meminit=False,
        # max_readers=1,
        map_size=int(10e12),
    )
    txn_write = env_new.begin(write=True)
    i = 0
    for index in tqdm(range(len(nmr_data))):
        inner_output = pickle.dumps(nmr_data[index], protocol=-1)
        txn_write.put(f'{i}'.encode("ascii"), inner_output)
        i += 1
        if i % 100 == 0:
            txn_write.commit()
            txn_write = env_new.begin(write=True)
    txn_write.commit()
    env_new.close()


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from collections import defaultdict
from unicore.data import BaseWrapperDataset, data_utils
from functools import lru_cache

class EpochResampleDataset(BaseWrapperDataset):
    def __init__(self, dataset, count_dict, seed, max_samples_per_class=None, min_samples_per_class=None):
        super().__init__(dataset)
        self.dataset = dataset
        self.count_dict = count_dict
        self.seed = seed
        self.max_samples_per_class = max_samples_per_class
        self.min_samples_per_class = min_samples_per_class
        self.filtered_indices = self.get_resample(1)

    def set_epoch(self, epoch):
        super().set_epoch(epoch)
        self.get_resample(epoch)

    def get_resample(self, epoch):
        with data_utils.numpy_seed(self.seed + epoch - 1):
            # count_dict = defaultdict(int)
            # for itr_class in self.sample_class:
            #     count_dict[itr_class] += 1
            filtered_indices = []
            for itr_class in self.count_dict.keys():
                # indices = [i for i, x in enumerate(self.sample_class) if x == itr_class]
                indices = self.count_dict[itr_class]
                if self.max_samples_per_class is not None:
                    if len(indices) > self.max_samples_per_class:
                        random_indices = np.random.choice(indices, self.max_samples_per_class, replace=False).tolist()
                        indices = random_indices
                if self.min_samples_per_class is not None:
                    if len(indices) < self.min_samples_per_class:
                        random_indices = np.random.choice(indices, self.min_samples_per_class).tolist()
                        indices = random_indices
                filtered_indices.extend(indices)
                # print("epoch", epoch, "class", itr_class, len(indices))
            self.filtered_indices = sorted(filtered_indices)
            # print("dataset", len(self.dataset), "self.filtered_indices", len(self.filtered_indices), self.filtered_indices[:10])
        
            return self.filtered_indices
        
    
    def __len__(self):
        return len(self.filtered_indices)

    @lru_cache(maxsize=16)
    def __getitem__(self, index: int):
        return self.dataset[self.filtered_indices[index]]

class EpochLogResampleDataset(BaseWrapperDataset):
    def __init__(self, dataset, count_dict, seed):
        super().__init__(dataset)
        self.dataset = dataset
        self.count_dict = count_dict
        self.seed = seed
        self.filtered_indices = self.get_resample(1)

    def set_epoch(self, epoch):
        super().set_epoch(epoch)
        self.get_resample(epoch)

    def get_resample(self, epoch):
        with data_utils.numpy_seed(self.seed + epoch - 1):
            # Calculate the log values and total log sum for each class
            log_counts = {cls: np.log(len(indices) + 1) for cls, indices in self.count_dict.items()}
            total_log_sum = sum(log_counts.values())

            # Calculate the resampling probability for each class
            resample_probs = {cls: log_counts[cls] / total_log_sum for cls in self.count_dict.keys()}
            
            # Calculate the total number of samples in the original dataset
            total_samples = sum(len(indices) for indices in self.count_dict.values())

            filtered_indices = []
            for itr_class, prob in resample_probs.items():
                indices = self.count_dict[itr_class]
                
                # Sample the number of samples based on log probability
                num_samples = int((total_log_sum) * prob)

                if len(indices) > num_samples:
                    random_indices = np.random.choice(indices, num_samples, replace=False).tolist()
                else:
                    random_indices = np.random.choice(indices, num_samples).tolist()

                filtered_indices.extend(random_indices)

            self.filtered_indices = sorted(filtered_indices)

            return self.filtered_indices
        
    def __len__(self):
        return len(self.filtered_indices)

    @lru_cache(maxsize=16)
    def __getitem__(self, index: int):
        return self.dataset[self.filtered_indices[index]]


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from functools import lru_cache
from unicore.data import BaseWrapperDataset


class RemoveHydrogenDataset(BaseWrapperDataset):
    def __init__(self, dataset, atoms, coordinates):
        self.dataset = dataset
        self.atoms = atoms
        self.coordinates = coordinates
        self.set_epoch(None)

    def set_epoch(self, epoch, **unused):
        super().set_epoch(epoch)
        self.epoch = epoch

    @lru_cache(maxsize=16)
    def __cached_item__(self, index: int, epoch: int):
        dd = self.dataset[index].copy()
        atoms = np.array(dd[self.atoms])
        coordinates = np.array(dd[self.coordinates])
        
        mask_hydrogen = atoms != 'H'
        atoms = atoms[mask_hydrogen]
        coordinates = coordinates[mask_hydrogen]

        dd[self.atoms] = atoms
        dd[self.coordinates] = coordinates.astype(np.float32)
        return dd

    def __getitem__(self, index: int):
        return self.__cached_item__(index, self.epoch)



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from functools import lru_cache
import logging
from unicore.data import BaseWrapperDataset, data_utils

logger = logging.getLogger(__name__)


class CroppingDataset(BaseWrapperDataset):
    def __init__(self, dataset, seed, atoms, coordinates, max_atoms=512):
        self.dataset = dataset
        self.seed = seed
        self.atoms = atoms
        self.coordinates = coordinates
        self.max_atoms = max_atoms   # max number of atoms in a molecule, None indicates no limit.
        self.set_epoch(None)

    def set_epoch(self, epoch, **unused):
        super().set_epoch(epoch)
        self.epoch = epoch

    @lru_cache(maxsize=16)
    def __cached_item__(self, index: int, epoch: int):
        dd = self.dataset[index].copy()
        atoms = dd[self.atoms]
        coordinates = dd[self.coordinates]
        if self.max_atoms and len(atoms) > self.max_atoms:
            with data_utils.numpy_seed(self.seed, epoch, index):
                index = np.random.choice(len(atoms), self.max_atoms, replace=False)
                atoms = np.array(atoms)[index]
                coordinates = coordinates[index]
        dd[self.atoms] = atoms
        dd[self.coordinates] = coordinates.astype(np.float32)
        return dd

    def __getitem__(self, index: int):
        return self.__cached_item__(index, self.epoch)


from fire import Fire

import warnings
import logging
import os

from ai2_kit.core.cmd import CmdGroup

logger = logging.getLogger(__name__)


class AlgorithmGroup:
    """
    Algorithms for specific domains.
    """

    def proton_transfer(self):
        """
        Proton transfer analysis toolkit.
        """
        from ai2_kit.algorithm import proton_transfer
        return proton_transfer.cli_entry

    def aosa(self):
        """
        Amorphous oxides structure analysis toolkit.
        """
        from ai2_kit.algorithm import aos_analysis
        return aos_analysis.cli_entry

    def reweighting(self):
        """
        Reweightning toolkit
        """
        from ai2_kit.algorithm.reweighting import ReweightingTool
        return ReweightingTool()


class WorkflowGroup:
    """
    Workflows for specific domains.
    """

    @property
    def cll_mlp_training(self):
        from ai2_kit.workflow.cll_mlp import run_workflow
        return run_workflow

    @property
    def fep_mlp_training(self):
        from ai2_kit.workflow.fep_mlp import run_workflow
        return run_workflow

class ToolGroup:
    """
    Tools for specific domains.
    """

    @property
    def ase(self):
        from ai2_kit.tool.ase import AseTool
        return AseTool

    @property
    def dpdata(self):
        from ai2_kit.tool.dpdata import DpdataTool
        return DpdataTool

    @property
    def model_devi(self):
        from ai2_kit.tool.model_devi import ModelDevi
        return ModelDevi

    @property
    def yaml(self):
        from ai2_kit.tool.yaml import YamlTool
        return YamlTool

    @property
    def batch(self):
        from ai2_kit.tool.batch import BatchTool
        return BatchTool

    @property
    def frame(self):
        from ai2_kit.tool.frame import FrameTool
        return FrameTool

    @property
    def hpc(self):
        # deprecated, use `oh-my-batch` instead
        warnings.warn('The hpc module is deprecated, please use oh-my-batch instead.',
                        DeprecationWarning, stacklevel=2)

        from ai2_kit.tool.hpc import cmd_entry
        return cmd_entry

    @property
    def lammps(self):
        from ai2_kit.tool.lammps import LammpsTool
        return LammpsTool()

    @property
    def misc(self):
        from ai2_kit.tool.misc import cmd_entry
        return cmd_entry

class FeatureGroup:
    """
    Featuring tools for specific domains.
    """

    @property
    def catalysis(self):
        """
        Catalyst specific tools.
        """
        from ai2_kit.feat.catalysis import CmdEntries
        return CmdEntries

    @property
    def cat(self):
        """
        Shortcut for catalyst.
        """
        return self.catalysis

    @property
    def spectr(self):
        """
        Spectrum specific tools.
        """
        from ai2_kit.feat.spectrum import CmdEntries
        return CmdEntries

    @property
    def nmrnet(self):
        """
        NMRNet specific tools.
        """
        try:
            from ai2_kit.algorithm.uninmr import CmdEntries
        except ImportError:
            logging.info('In order to use nmrnet, you need to ensure the following packages are installed: ')
            logging.info('"rdkit", "scipy" and "unicore"')
            raise
        return CmdEntries


ai2_kit = CmdGroup({
    'workflow': WorkflowGroup(),
    'algorithm': AlgorithmGroup(),
    'tool': ToolGroup(),
    'feat': FeatureGroup(),

}, doc="Welcome to use ai2-kit!")


def _setup_logging():
    level_name = os.environ.get('LOG_LEVEL', 'INFO')
    level = logging._nameToLevel.get(level_name, logging.INFO)
    logging.basicConfig(format='%(asctime)s %(name)s: %(message)s', level=level)
    logging.getLogger('transitions.core').setLevel(logging.WARNING)


def main():
    _setup_logging()
    Fire(ai2_kit)


def rpc_main():
    from fire_rpc import make_fire_cmd
    from ai2_kit.core.util import json_dumps
    _setup_logging()
    Fire(make_fire_cmd(ai2_kit, json_dumps=json_dumps))


if __name__ == '__main__':
    main()



from .unimat import UniMatModel
from .transformer_encoder_with_pair import TransformerEncoderWithPair
# from .unimof_v2 import UniMOFV2Model
# from .unimat import UniMatModel
# from .unimof_v2_NoGasID import UniMOFV2Model_NoGasID